//  Mini STEM/Game Platform (Nano ver) Library - Hardware Support functions
//  4 Button module + 8 LED module + speaker out (piezo)
// 
// license: Attribution-NonCommercial-ShareAlike
// 2019-2020 Ron Miller, Initial Development
// see: https://www.instructables.com/id/Mini-STEM-LED-Game-Platform/
//
// 2023 also supports RP2040

// ----------------  Function prototypes (pre definitions)
void gPlatform_setup();

void refreshDisp();
void refreshWait(int);
void set_LEDs(byte, byte);
void clearDisp();

void scanBtns();
bool btnPressed();

void beep(int, int);
void boop(int, int);
void buZZ(int, int);

//=================================================================
// Edit within this section if your device is not wired up as a
// standard "Mini STEM/Game Platform" see: instructables.com/....
// ----------------  hardware dependent section  ------------------

//===========================================================================================================
//=========== Below unComment with "/*****/" the section which applies & otherwise disable ( with: "/*****" )

/*****/
// ------------ Arduino Nano open prototype STEM Educational platform version
//                   (could be on a breadboard, top-side of PCB or otherwise directly attached)
const int button[] = {16, 17, 18, 19}; // The four button input pins are oriented in this order
const int lites[] = {9, 8, 7, 6, 5, 4, 3, 2};     // the LED pin assignments
/*****
// ---------- Arduino Nano finished enclosed version (also aka Left-hand version, all reversed right to left)
//            3D enCased STEM Prj 1 2 or 3
const int button[] = {19, 18, 17, 16};        //The four button input pins
const int lites[] = {2, 3, 4, 5, 6, 7, 8, 9};     //the LED pins
/*****
// ------------ Plug-N-Play (held together with just an alastic band) - - - - - - - - - - WaveShare RP2040-Zero
//                          (could be on a breadboard, top-side of PCB, or otherwise directly attached)
#define RP2040
const int button[] = {27, 26, 15, 14};     //NOTE: button input pins need pullup Rs
const int lites[] = {1, 2, 3, 4, 5, 6, 7, 8};     //the LED pins
/*****
// ------------ enCased STEM Prj 1 2 or 3  with WaveShare RP2040-Zero
#define RP2040
const int button[] = {14, 15, 26, 27};     //NOTE: button input pins need pullup Rs
const int lites[] = {8, 7, 6, 5, 4, 3, 2, 1};     //the LED pins
/*****/
//===========================================================================================================


#ifdef RP2040
  const byte AUDIO_OUT = 12;
  const byte AUDIO_NEG = 10;
  #define PSUDOVCC 0    // supplies +V to 8 LED module
  #define PSUDOGND 28   // used by 4 Key module as needed
  const int ledOnBrd=16;  // for the WS2812 LED of Waveshare RP2040-Zero
#else
  const byte AUDIO_OUT = 13;
  const byte AUDIO_NEG = 12;
  #define PSUDOVCC 10  // supplies +V to 8 LED module
  #define PSUDOGND 15  // (aka A1) used by 4 Key module as needed
  const int ledOnBrd=LED_BUILTIN;  // the number of the LED 'L' on MCU board
#endif

// constants won't change, unless the hardware changes.
// ---------- 8 LED configuration data
const int nleds=8;
#define ON_STATE 0
// ----------  button config. data
#define PRESSED_STATE 0

bool Loud = false;  // enables AC driven audio

//=================================================================

const bool debugPrt=false;
// ------------------------------------------------------------
// 4-Button module support

bool btn1 = false;  // reflects the state of button #1
bool btn2 = false;
bool btn3 = false;
bool btn4 = false;
bool Btn=false;   // reflects Buttons collectively
bool btn[5];      // array of btn1-4 states
bool btns=false;  // true is multiple buttons are pressed
byte btnNum;     // btn number pressed 1-4, lower num takes precedence
bool ESC = false;  // a psuedo button, true is btn4 is held down for 2 seconds

bool btn1Changed = false; // true if btn1 just changed
bool btn2Changed = false;
bool btn3Changed = false;
bool btn4Changed = false;
bool BtnChanged = false;  // true if any button changes
unsigned long lastChg;    // millis() when last button change occured
long debounceDelay = 2;    // the debounce time; increase if button chatters
bool btnStable=false;
byte chgCnt;

#define btn1Pressed (btn1Changed && btn1)
#define btn2Pressed (btn2Changed && btn2)
#define btn3Pressed (btn3Changed && btn3)
#define btn4Pressed (btn4Changed && btn4)
#define BtnPressed (BtnChanged && Btn)

// ------------------------------------------------------------
// 8 LED display bar support
unsigned int msCnt;   // pseudo MSecs, incremented every time 'refreshDisp()' is called
// NOTE! in the arrays below, elements [1]-[8] relate to LED 1-8 respectively
//                            [0] & [9] are for overflow protection
bool dim[8+2];  // set true which LEDs you want to be dim (~10% on)
bool lit[8+2];  //                 "                  lit (~33% on)
bool brt[8+2];  //                 "                  bright (100%)
int brightOne;  // sets indicate (1-8) LED Bright.  Useful as a cursor or sprite
int litLED;     //         "               Lit            "
int dimLED=0;
int flashLed;
int dimFlash;
byte ledsOn;  // the # of LEDs current On
int darkCnt;  // track time with dark display & no input
int spriteOne, spriteTwo;
bool onSet;   // if true LEDs set as is, upd disabled

//==================== Support functions ========================

// ----------------- setup the hardware for the Mini STEM/Game platform
void gPlatform_setup() {
  // ------------ setup LED pins as outputs
  for (int x = 0; x < nleds; x++)
  {
    pinMode(lites[x], OUTPUT);
//    digitalWrite(lites[x], ON_STATE);  // on startup have LEDs momentarially On
    digitalWrite(lites[x], !ON_STATE);  // nah, now dis the blib
  }
//  pinMode(PSUDOVCC, OUTPUT);    // provide +v for LED bar module
//  digitalWrite(PSUDOVCC, HIGH);
  pinMode(PSUDOVCC, INPUT);   // let it be externally driven for my blackBox proj.

  // ------------ setup the four button module
  for (int x = 0; x < 4; x++)
  {
    pinMode(button[x], INPUT_PULLUP);
  }
  pinMode(PSUDOGND, OUTPUT);    // provide Ground (-V) for the button module
  digitalWrite(PSUDOGND, LOW);

  // ------------ setup and init the Audio
  pinMode(AUDIO_OUT, OUTPUT);
  digitalWrite(AUDIO_OUT, LOW);
  pinMode(AUDIO_NEG, OUTPUT);
  digitalWrite(AUDIO_NEG, LOW);

  clearDisp();  // initialize display LEDs
  refreshDisp();
  scanBtns();   //   "  button statuses
}

// ------------------------------------------------------------
//        Display Output Processing

//  ---------------------  Display Update  - Should to be called ~ every One-two Milli-Seconds
void refreshDisp() {
  bool on[13];
  int ledn, Lite;
  unsigned int ledTime;
  bool flashTime, dTime;
  int i, n;
  
  msCnt++;
  if (onSet) return;   // override: LEDs are set in fixed On pattern
  for (i=1; i<=nleds; i++) on[i]=0; // init working array

  // -------------- determine which Leds its time to illuminate
  ledTime = msCnt;
//  dTime = false;
  if (ledTime%15 == 0) {                // 1/x cadence time to show 'dim'
//    dTime = true;
    for (i=1; i<=nleds; i++) {on[i]=dim[i];}
    on[dimLED] = true;
  } else if (ledTime%3 == 1) {          // 33% of the time process 'lit'
    for (i=1; i<=nleds; i++) on[i]=lit[i];
    on[litLED] = true;
  }

  for (i=1; i<=nleds; i++) on[i]|=brt[i];
  on[brightOne] = true;
  on[spriteOne] = true;
  on[spriteTwo] = true;
  flashTime = ((msCnt%500) < 200);  // 200ms out of 500ms turn it On
  on[flashLed] = flashTime;
  dTime=ledTime%13 == 6;  // do Dim flash at dim freq, out of phase
  if (dTime) on[dimFlash]=flashTime;
  
  // -------------- drive the resulting Leds of interest
  ledsOn=0;
  for (i=0; i<nleds; i++) {
    ledn = i+1;
    Lite = lites[i];
    digitalWrite(Lite, (on[ledn]==ON_STATE));
    if (on[ledn]) ledsOn++;
  }
}

// -----------------------
void refreshWait(int msec) {
  do {
    refreshDisp();
    if (msec>0) delay(1);
    msec--;
  } while (msec>0);
}

// -----------------------   
void showColor(byte color) {
  if (color==0) clearDisp();
  if (color==1 || color==5)  lit[1] = lit[2] = true; // set 1 LEDs
  if (color==2 || color==5)  lit[3] = lit[4] = true; // set 2 LEDs
  if (color==3 || color==5)  lit[5] = lit[6] = true; // set 3 LEDs
  if (color==4 || color==5)  lit[7] = lit[8] = true; // set 4 LEDs
}

void set_LEDs(byte led_pattern, byte state) {
  // led_pattern: bit decoded pattern, eg: 9= (00001001b)left most led + forth (led#1 & #4)
  // state: 1=dim, 2=lit, 3=bright, 0=clear
  byte b;
  for (int i=1; i<=8; i++) {
    b = led_pattern & 1;
    led_pattern=led_pattern>>1;
    if (b) {
      dim[i]=0; lit[i]=0; brt[i]=0; // default state cleared
      if (state==1)  dim[i] = true;
      else if (state==2)  lit[i] = true;
      else if (state==3)  brt[i] = true;
    }
  }
  if (state==0 || state==3) refreshDisp();  // reflect hard ONs or OFFs now
}

// ------------------
void clearDisp() {
    onSet=false;
    dimLED=0;
    litLED=0;
    brightOne = 0;
    flashLed = 0;
    dimFlash = 0;
    spriteOne = 0;
    spriteTwo = 0;
    for (int i=0; i<=nleds; i++) {
      dim[i] =false;
      lit[i]=false;
      brt[i]=false;
    }
    if (!onSet) refreshDisp();
}

// ------------------------------------------------------------
//        Button Input Processing

void scanBtns() {
  static byte btnState = 0, currState, priorBtn;
  bool b1,b2,b3,b4;
  byte i,k;

  darkCnt = (ledsOn==0)? darkCnt+1 : 0;
  if (darkCnt>30000) {  // nothing happing ... Atention Alert !
    digitalWrite(lites[0], ON_STATE);  digitalWrite(lites[7], ON_STATE);
    beep(25,50);
    digitalWrite(lites[0], !ON_STATE);  digitalWrite(lites[7], !ON_STATE);
    darkCnt=0;
  }
  //if ((millis()%1000)==1) {Serial.print("  darkCnt: "); Serial.println(darkCnt);}
  
  BtnChanged = btn1Changed = btn2Changed = btn3Changed = btn4Changed = false;
  b1 = btn[1] = !digitalRead(button[0]);
  b2 = btn[2] = !digitalRead(button[1]);
  b3 = btn[3] = !digitalRead(button[2]);
  b4 = btn[4] = !digitalRead(button[3]);

  currState = int(b1) + 2*int(b2) + 4*int(b3) + 8*int(b4);  // combined button states
  if (currState != btnState) {
    btnStable=false;
    if(++chgCnt < debounceDelay) return;
  }
  btnStable=true;
  chgCnt = 0;
  Btn = b1 || b2 || b3 || b4;
  btns = (b1 + b2 + b3 + b4) > 1;

  if (debugPrt)  {Serial.print("Btns:  ");  Serial.print(b1); Serial.print(b2); Serial.print(b3); Serial.print(b4); }
  btnNum =  (b1)? 1 : ((b2)? 2 : ((b3)? 3 : ((b4)? 4 : 0)));  // logical button # (1-4) (multiple presses not supported)
  
  if (currState != btnState) {  // the 'button state' has changed
    darkCnt=0;
    if (debugPrt)  {Serial.print("CurrState Btn State:  ");  Serial.println(currState);}
    lastChg = millis();
    priorBtn = btnState;
    btnState = currState;
    BtnChanged=true;

    Btn = b1 || b2 || b3 || b4;
    if(b1 != btn1) {btn1 = b1; btn1Changed = true;}
    if(b2 != btn2) {btn2 = b2; btn2Changed = true;}
    if(b3 != btn3) {btn3 = b3; btn3Changed = true;}
    if(b4 != btn4) {btn4 = b4; btn4Changed = true;}
  }
  
  if (btn4) ESC = (millis()-lastChg)>2000;
  else ESC = false;
  if (btn1 && btn4) ESC = true;  // will not be ok if b1+4 is valid in current activity
}

// -----------------------   
bool btnPressed() { // check for a new button pressing
  scanBtns();
  return (BtnChanged && Btn!=0);
}

// ------------------------------------------------------------
//        Audio support

void beep(int msecs, int delayms=0){  // works OK with Buzzers, best with Speakers
  for (int i=0; i<msecs; i++) {         // loop for so many 'msecs'
    digitalWrite ( AUDIO_OUT, HIGH); 
    delayMicroseconds(500);    
    digitalWrite ( AUDIO_OUT, LOW);
    delayMicroseconds(500);
    refreshDisp();  // just spent 1 ms, lets refreshDisp now
  }
  refreshWait(delayms);          // pause for delayms millisecs   
}

void boop(int msecs, int delayms=0){   // produces a lower frequency tone then does 'beep()'
  for (int i=0; i<((msecs+3)/4); i++) {         // loop for 'msecs' milliseconds
    digitalWrite ( AUDIO_OUT, HIGH); 
    refreshWait(2);
    digitalWrite ( AUDIO_OUT, LOW);
    refreshWait(2);
  }
  refreshWait(delayms);          // pause for delayms millisecs
}

/***
void buZZ(int msecs, int delayms=0){   // produces a lower frequency tone then does 'beep()'
  for (int i=0; i<((msecs+5)/6); i++) {         // loop for 'msecs' milliseconds
    beep(2,1);
    boop(2);
  }
  refreshWait(delayms);          // pause for delayms millisecs
}
***/

void buZZ(int msecs, int delayms=0) {   //produces a aggregated beep sound
  for (int i=0; i<msecs; i+=5) {       // wait for >= delayms ms
    beep(2);
    digitalWrite ( AUDIO_OUT, HIGH);
    delayMicroseconds(1500);
    digitalWrite ( AUDIO_OUT, LOW);
    delayMicroseconds(1500);
  }
  //digitalWrite ( AUDIO_OUT, LOW);
  refreshWait(delayms);          // pause for delayms millisecs
}  
